import { BaseReader }               from '../readers/base.reader';
import { IReaderConfiguration }     from "../models/reader.configuration";
import { IBarcodeValue }            from '../models/ibarcode-value';
import { ApplicationIdentifier }    from '../models/application-identifier';
import { BarcodeValue }             from '../models/barcode-value';
import { APPLICATION_IDENTIFIERS }  from '../models/application-identifiers';
import { CONTROL_CHARS }            from '../config/control-characters.config';
import { AIM_CODES }                from '../config/aim-codes';
import { AimParser }                from '../utils/aim-parser';


const SYMBOLOGY: string = 'gs1_128';
const DELIMITER: string = ' ';

export class GS1Reader extends BaseReader {

	constructor(readerConfig?: IReaderConfiguration) {
		super(SYMBOLOGY, null, readerConfig);
	}

	protected removeControlCharacters(value: string): string {
		let result = value;
		CONTROL_CHARS.forEach((charCode) => {
			result = result.replace(String.fromCharCode(charCode), DELIMITER);
		});
		return result;
	}

	public validate(value: string): boolean {
		let idPrefix = value.indexOf(AIM_CODES.GS1);
		let valueLength = AimParser.parseAimCode(this.symbology, value).length;
		return idPrefix === 0 && valueLength > 0;
	}

	decode(value: string): IBarcodeValue {
		let sterilizedValue = this.removeControlCharacters(value);
		let result = new BarcodeValue(this.symbology, value);
		try {
			this.tryValidate(value);
			let valWithoutId = AimParser.parseAimCode(this.symbology, sterilizedValue);
			result.values = [].concat.apply([], valWithoutId.split(DELIMITER).map((val) => this.parseValues(val)));
		} catch (ex) {
			result.success = false;
			result.errorMessage = ex;
		}
		return result;
	}

	protected findAi(value: string): ApplicationIdentifier {
		let ai: ApplicationIdentifier = null;
		let codeLength = 2;
		while (ai === null && codeLength < 5) {
			let code = value.substr(0, codeLength);
			let ais = APPLICATION_IDENTIFIERS.filter((x) => {
				return x.code === code;
			});
			if (ais.length > 0) {
				ai = ais[0];
			} else {
				codeLength++;
			}
		}
		return ai;
	}

	protected parseValue(ai: ApplicationIdentifier, input: string): Object {
		let val: any = null;
		if (ai.fractional !== true) {
			val = input.substr(ai.code.length, ai.length);
		} else {
			let scale = Number(input.charAt(ai.code.length)),
				num = input.substr(ai.code.length + 1, ai.length);

			val = Number(num) / Math.pow(10, scale);
		}

		return {
			code: ai.code,
			value: val
		};
	}

	protected parseValues(input: string): Object[] {
		let vals = Array<Object>(),
			ai = this.findAi(input);

		if (input.length > ai.totalLength) {
			vals = vals.concat(this.parseValues(input.substr(ai.totalLength)));
		}

		vals.push(this.parseValue(ai, input));

		return vals;
	}
}
